package falkore.core;

import java.awt.Dimension;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.event.KeyEvent;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;

import falkor.model.GameObject;


public class GameState {

   public enum State{ RUN, STOP };
   public enum RunState{ GREET, GAME, PAUSE, HISCORE };

   public static int GAME_AREA_WIDTH = 512;
   public static int GAME_AREA_HEIGHT = 316;
   
   private Map<Integer, Boolean> keypressMap;
   private Map<Integer, Boolean> mouseMap;

   private Point mousePosition;
   private Dimension gameArea;

   private Collection<GameObject> gameObjects;

   private Integer mouseWheelDelta;
   private Integer lastPause;
   private Integer renderDelay;
   private Integer logicDelay;
   private Integer startupTime;
   private Integer logicTicks, renderTicks;

   private State state;
   private RunState runState;

   public GameState() {
      reset();
   }

   public String toString() {
      return "keys: "+keypressMap+
            "\n mouse buttons: "+mouseMap+
            "\n mouse pos: "+mousePosition+
            "\n mouse wheel: "+mouseWheelDelta+
            " render delay: "+renderDelay+
            " logic delay: "+logicDelay+
            " state: "+state+
            " runState: "+runState+
            "\n startupTime: "+startupTime+
            " logicTicks: "+logicTicks+
            " renderTicks: "+renderTicks+
            "\n gameArea: "+gameArea;
   }

   private void renderTick() {
      ++renderTicks;
   }

   private void logicTick() {
      ++logicTicks;
   }

   public void reset() {
      renderDelay = 16;
      logicDelay = 8;
      startupTime = (int) System.currentTimeMillis();
      gameArea = new Dimension(GAME_AREA_WIDTH, GAME_AREA_HEIGHT);
      state = State.RUN;
      runState = RunState.GREET;
      lastPause = -1;
      logicTicks = 0;
      renderTicks = 0;

      mouseWheelDelta = 0;
      mousePosition = new Point(0,0);

      gameObjects = new HashSet<GameObject>();
      keypressMap = new HashMap<Integer, Boolean>();
      mouseMap = new HashMap<Integer, Boolean>();
   }

   public void step() {
      long frameStart, frameDuration;
      int msToSleep;

      frameStart = getRunningTime();

      System.out.println(toString());

      /* Enforce state transitions */
      if(isDown(KeyEvent.VK_ESCAPE)) {
         stop();
      } else if(isDown(KeyEvent.VK_P)) {
         togglePause();
      } else if(runState == RunState.GREET && isDown(KeyEvent.VK_SPACE) ) {
         start();
      } else if(runState == RunState.GAME && isDown(KeyEvent.VK_K)) {
         highScore();
      } else if(runState == RunState.HISCORE && isDown(KeyEvent.VK_SPACE)) {
         reset();
      } else {

         /* Process the current game state */
         switch(runState){
         case GREET:

            break;
         case GAME:

            break;
         case PAUSE:

            break;
         case HISCORE:

            break;
         }
      }

      frameDuration = getRunningTime() - frameStart;
      msToSleep = getLogicDelay() - msToLogicTicks(frameDuration);
      if( msToSleep > 0 ) {
         try {
            Thread.sleep(msToSleep);
         } catch (InterruptedException ie) {
            ie.printStackTrace();
            /* TODO: handle this better */
         }
      } else {
         System.out.println("warning: state step exceeded time quantum!");
      }

      logicTick();

   }

   public void render(Graphics2D gfx) {
      long frameStart, drawDuration;
      int msToSleep;
      frameStart = getRunningTime();

      for( GameObject obj : gameObjects ) {
         obj.draw(gfx);
      }

      drawDuration = getRunningTime() - frameStart;
      msToSleep = getRenderDelay() - msToRenderTicks(drawDuration);

      if( msToSleep > 0 ) {
         try {
            Thread.sleep(msToSleep);
         } catch (InterruptedException ie) {
            ie.printStackTrace();
         }
      } else {
         System.out.println("warning: render exceeded time quantum!");
      }

      renderTick();
   }

   private long getRunningTime() {
      return startupTime - System.currentTimeMillis(); 
   }

   public void pause() {
      if( runState != RunState.PAUSE ) {
         togglePause();
      }
   }

   public void updateKeypressMap( int key, boolean down ) {
      keypressMap.put(key, down);
   }

   public void updateMousePosition( int x, int y ) {
      mousePosition.x = x;
      mousePosition.y = y;
   }

   public void updateMouseMap( int button, boolean down ) {
      mouseMap.put(button, down);
   }

   public void updateMouseWheelDelta( int delta ) {
      mouseWheelDelta += delta;
   }

   private boolean isDown( int key ) {
      if( keypressMap.containsKey(key) && keypressMap.get(key ) ) {
         return true;
      } else {
         return false;
      }
   }

   private void start() {
      runState = RunState.GAME;
   }

   private void stop() {
      state = State.STOP;
   }

   private void togglePause() {
      /* Ignore pause toggles that occurred in the last 30 logic ticks. 
       * Note that 30 * 8 (the current logic delay) is 240 milliseconds,
       * or slightly less than a quarter of a second.  
       */
      if ( logicTicks - lastPause < 30 ) {
         return;
      }

      if( runState == RunState.PAUSE ) {
         runState = RunState.GAME;
      } else if( runState == RunState.GAME ){
         runState = RunState.PAUSE;
      }

      lastPause = logicTicks;
   }

   private void highScore() {
      runState = RunState.HISCORE;
   }

   private int msToLogicTicks( long ms ) {
      return (int) ( (double)ms/getLogicDelay() );
   }

   private int getLogicDelay() {
      return logicDelay;
   }

   private int msToRenderTicks( long ms ) {
      return (int) ( (double)ms/getRenderDelay() );
   }

   private int getRenderDelay() {
      return renderDelay;
   }

   public GameState.State getState() {
      return state;
   }

   public GameState.RunState getRunState() {
      return runState;
   }
  
   public Dimension getGameArea() {
      return gameArea;
   }
}
